﻿using System;
using System.IO;
using System.Collections;

namespace BeepMap {
    partial class WaveGen {
        ISndPlay sndOut;
        int sampleRate = 44100;
        float a4freq = 440;
        int waveLength = -1;
        short[] mixedWave;

        public enum Waveform {
            Sine,
            SinePow2,
            SinePow4,
            Square5,
            Square10,
            Square15,
            Square20,
            Square25,
            Square30,
            Square35,
            Square40,
            Square45,
            Square50,
            Triangle,
            Sawtooth,
            WhiteNoise
        }

        public WaveGen(ISndPlay OutputObject) {
            sndOut = OutputObject;
        }

        public int SetRate {
            get { return sampleRate; }
            set { sampleRate = value; }
        }

        public void ExportWAV(string filename) {
            FileStream fs = new FileStream(filename, FileMode.Create);
            ExportWAV(fs);
            fs.Close();
        }

        public void ExportWAV(Stream fileStream) {
            BinaryWriter bw = new BinaryWriter(fileStream);

            // write RIFF header
            bw.Write("RIFF".ToCharArray());

            // file size (data chunk size + 36)
            bw.Write(Convert.ToUInt32((waveLength*2) + 36));

            // write WAVE header
            bw.Write("WAVE".ToCharArray());

            // format chunk marker
            bw.Write("fmt ".ToCharArray());

            // sample length
            bw.Write(Convert.ToUInt32(16));

            // wave type (0x01 = PCM)
            bw.Write(Convert.ToUInt16(1));

            // number of channels (1 = mono, 2 = stereo, etc)
            bw.Write(Convert.ToUInt16(1));

            // sample rate
            bw.Write(Convert.ToUInt32(44100));

            // bitrate
            bw.Write(Convert.ToUInt32(88200));

            // block align
            bw.Write(Convert.ToUInt16(2));

            // bits per sample
            bw.Write(Convert.ToUInt16(16));

            // data chunk header
            bw.Write("data".ToCharArray());

            // size of data section
            bw.Write(Convert.ToUInt32(waveLength*2));

            // write data
            for(int i=0; i < mixedWave.Length; i++)
                bw.Write(Convert.ToInt16(mixedWave[i]));
            
            // close stream
            bw.Close();
        }

        public void Play() {
            if(waveLength > 0)
                sndOut.Play(sampleRate, waveLength, mixedWave);
        }

        public void Stop() {
            sndOut.Stop();
        }

        public void RenderPattern(Channel chan, Pattern pattern, decimal BPM, decimal A4Frequency) {
            int wavePtr = 0;

            // determine how many samples will be in the final wave and generate scratch wave
            int numSamples = 0;
            foreach(Note note in pattern.notes) {
                numSamples += CalculateSamples(note.notevalue, BPM);
            }

            waveLength = numSamples;
            double[] mixedWaveDouble = new double[numSamples];

            // step through notes and build wave
            foreach(Note note in pattern.notes) {
                // don't render anything if the current note is a rest
                if(!note.sleep) {
                    float frequency = CalculateNoteFrequency(a4freq, note.distance);
                    int samples = CalculateSamples(note.notevalue, BPM);
                    double[] wave = SelectGenerator(frequency, samples, ((float)chan.volume * 0.01f), chan.waveform);

                    // run the wave through A/D filters if specified
                    if(chan.attack > 0) wave = FilterAttack(wave, chan.attack);
                    if(chan.decay > 0) wave = FilterDecay(wave, chan.decay);

                    // copy wave to final stream
                    wave.CopyTo(mixedWaveDouble, wavePtr);
                }

                // advance wave pointer
                wavePtr += CalculateSamples(note.notevalue, BPM);
            }

            // convert the resulting double to a 16-bit integer
            mixedWave = new short[numSamples];

            for(int i=0; i < numSamples; i++)
                mixedWave[i] = Convert.ToInt16((double)mixedWaveDouble[i] * Int16.MaxValue); ;
        }

        private double[] RenderChannel(Channel chan, decimal bpm) {
            int wavePtr = 0;

            // determine how many samples will be in the final wave and generate scratch wave
            int numSamples = 0;
            foreach(int orderVal in chan.order) {
                Pattern pattern = (Pattern)chan.patterns[orderVal];
                foreach(Note note in pattern.notes) {
                    numSamples += CalculateSamples(note.notevalue, bpm);
                }
            }
            double[] chanSamples = new double[numSamples];

            // step through notes and build wave
            foreach(int orderVal in chan.order) {
                Pattern pattern = (Pattern)chan.patterns[orderVal];
                foreach(Note note in pattern.notes) {
                    // don't render anything if the current note is a rest
                    if(!note.sleep) {
                        float frequency = CalculateNoteFrequency(a4freq, note.distance);
                        int samples = CalculateSamples(note.notevalue, bpm);
                        double[] wave = SelectGenerator(frequency, samples, ((float)chan.volume * 0.01f), chan.waveform);

                        // run the wave through A/D filters if specified
                        if(chan.attack > 0) wave = FilterAttack(wave, chan.attack);
                        if(chan.decay > 0) wave = FilterDecay(wave, chan.decay);

                        // copy wave to final stream
                        wave.CopyTo(chanSamples, wavePtr);
                    }

                    // advance wave pointer
                    wavePtr += CalculateSamples(note.notevalue, bpm);
                }
            }

            // done!
            return chanSamples;
        }

        private double[] SelectGenerator(float frequency, int samples, float volume, Waveform waveform) {
            switch(waveform) {
                case Waveform.Sine:
                    return WaveSine(frequency, samples, volume, 1);
                case Waveform.SinePow2:
                    return WaveSine(frequency, samples, volume, 2);
                case Waveform.SinePow4:
                    return WaveSine(frequency, samples, volume, 4);
                case Waveform.Square5:
                    return WaveSquare(frequency, samples, volume, 0.05);
                case Waveform.Square10:
                    return WaveSquare(frequency, samples, volume, 0.10);
                case Waveform.Square15:
                    return WaveSquare(frequency, samples, volume, 0.15);
                case Waveform.Square20:
                    return WaveSquare(frequency, samples, volume, 0.20);
                case Waveform.Square25:
                    return WaveSquare(frequency, samples, volume, 0.25);
                case Waveform.Square30:
                    return WaveSquare(frequency, samples, volume, 0.30);
                case Waveform.Square35:
                    return WaveSquare(frequency, samples, volume, 0.35);
                case Waveform.Square40:
                    return WaveSquare(frequency, samples, volume, 0.40);
                case Waveform.Square45:
                    return WaveSquare(frequency, samples, volume, 0.45);
                case Waveform.Square50:
                    return WaveSquare(frequency, samples, volume, 0.50);
                case Waveform.Triangle:
                    return WaveTriangle(frequency, samples, volume);
                case Waveform.Sawtooth:
                    return WaveSawtooth(frequency, samples, volume);
                case Waveform.WhiteNoise:
                    return WaveWhiteNoise(samples, volume);
            }

            return null;
        }

        public void Render(ArrayList Channels, decimal BPM, decimal A4Frequency) {
            int totalSamples = 0;

            // generate the waves
            a4freq = Convert.ToSingle(A4Frequency);
            int numChannels = Channels.Count;
            double[][] waves = new double[numChannels][];

            for(int i=0; i < numChannels; i++) {
                waves[i] = RenderChannel((Channel)Channels[i], BPM);
            }

            // determine how long the mixed wave should be and create the mixdown output
            for(int i=0; i < numChannels; i++)
                if(waves[i].Length > totalSamples) totalSamples = waves[i].Length;

            double[] mixedWaveDouble = new double[totalSamples];
            waveLength = totalSamples;

            // mix the waves :o
            for(int j=0; j < numChannels; j++) {
                for(int i=0; i < waves[j].Length; i++) {
                    double mixed = mixedWaveDouble[i] + waves[j][i];
                    mixed = (mixed > 1) ? 1 : mixed;    // CLIPPING DETECTED!!!
                    mixed = (mixed < -1) ? -1 : mixed;  // OH NOES :O

                    mixedWaveDouble[i] = mixed;
                }
            }

            // convert the resulting double to a 16-bit integer
            mixedWave = new short[totalSamples];

            for(int i=0; i < totalSamples; i++)
                mixedWave[i] = Convert.ToInt16(mixedWaveDouble[i] * Int16.MaxValue); ;

            // done!
        }

        public float CalculateMilliseconds(float notevalue, decimal bpm) {
            return (60f / Convert.ToSingle(bpm)) * notevalue * 1000;
        }

        public int CalculateSamples(float notevalue, decimal bpm) {
            return (int)(sampleRate * ((float)CalculateMilliseconds(notevalue, bpm) / 1000));
        }

        public float CalculateNoteFrequency(float a4hertz, int distance) {
            return (float)System.Math.Pow(2.0f,((float)distance / 12.0f)) * a4hertz;
        }
    }

    public struct Note {
        public float notevalue;
        public int distance;
        public bool sleep;

        public override string ToString() {
            if(sleep)
                return "(Rest) " + notevalue;
            else
                return distance + " " + notevalue;
        }
    }

    public struct Pattern {
        public int id;
        public ArrayList notes;

        public override string ToString() {
            return id.ToString();
        }
    }

    public struct Channel {
        //public ArrayList notes;
        public ArrayList patterns;
        public ArrayList order;
        internal WaveGen.Waveform waveform;
        public byte volume;
        public string name;
        public decimal attack;
        public decimal decay;

        public int GetNextFreeID() {
            int foundID = -1;
            for(int i=0; i < int.MaxValue; i++) {
                foundID = i;
                foreach(Pattern pat in patterns) {
                    if(pat.id == i) {
                        foundID = -1;
                        break;
                    }
                }

                if(foundID >= 0) break;
            }
            
            return foundID;
        }

        public void RemovePattern(int index) {
            for(int i=order.Count; i > -1; i--)
                if(Convert.ToInt32(order[i]) == index)
                    order.RemoveAt(i);
        }

        public void SortPatterns() {
            // I hate implementing sorting algorithms, so here's a stupid and lazy one
            int total = patterns.Count, numSorted = 0, currentID = 0;
            ArrayList newPatterns = new ArrayList();

            while(numSorted < total) {
                foreach(Pattern pat in patterns) {
                    if(pat.id == currentID) {
                        newPatterns.Add(pat);
                        numSorted++;
                        break;
                    }
                }
                currentID++;
            }

            patterns = newPatterns;
        }

        public override string ToString() {
            return name;
        }
    }
}
